@@ -44,6 +44,22 @@ int max_int(int a, int b)
 	return b;
 }
 
+int min_int(int a, int b)
+{
+	if (a >= b) {
+		return b;
+	}
+	return a;
+}
+
+uint64_t min_uint64_t(uint64_t a, uint64_t b)
+{
+	if (a >= b) {
+		return b;
+	}
+	return a;
+}
+
 void enforce_range(const char *name, int v, int min, int max)
 {
 	if (check_range(v, min, max) == EXIT_FAILURE) {
@@ -254,7 +270,7 @@ int file_exists(char *name)
 #include <uuid/uuid.h>
 #endif
 
-int drop_privs()
+int drop_privs(void)
 {
 	struct passwd *pw;
 	if (geteuid() != 0) {
